# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from iconsdk.utils.convert_type import convert_params_value_to_hex_str
from iconsdk.exception import DataTypeException


class Call:
    """
    Class `Call` for calling a SCORE API.
    Once an instance generated, it is read-only."""
    def __init__(self, from_: str, to: str, method: str, params: dict):
        self.__from = from_
        self.__to = to
        self.__method = method
        self.__params = params

    @property
    def from_(self):
        return self.__from

    @property
    def to(self):
        return self.__to

    @property
    def method(self):
        return self.__method

    @property
    def params(self):
        return convert_params_value_to_hex_str(self.__params) if self.__params else None

    def to_dict(self):
        return {"from_": self.from_, "to": self.to, "method": self.method, "params": self.params}


class CallBuilder:
    """
    Builder for a `Call` object.
    Once setting it, a value of any property can't be changed forever.
    """
    def __init__(self, from_: str=None, to: str=None, method: str=None, params: dict=None) -> Call:
        self._from_ = from_
        self._to = to
        self._method = method
        self._params = params

    def from_(self, from_):
        self._from_ = from_
        return self

    def to(self, to):
        self._to = to
        return self

    def method(self, method):
        self._method = method
        return self

    def params(self, params):
        self._params = params
        return self

    def build(self) -> Call:
        return Call(self._from_, self._to, self._method, self._params)

    @classmethod
    def from_dict(cls, call_as_dict: dict) -> 'CallBuilder':
        """Returns a CallBuilder made from dict"""
        try:
            return cls(
                from_=call_as_dict['from_'],
                to=call_as_dict['to'],
                method=call_as_dict['method'],
                params=call_as_dict['params'] if "params" in call_as_dict else None
            )
        except KeyError:
            raise DataTypeException("The input data invalid. Mapping key not found.")





# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hashlib import sha3_256
from copy import deepcopy


translator = str.maketrans({
    "\\": "\\\\",
    "{": "\\{",
    "}": "\\}",
    "[": "\\[",
    "]": "\\]",
    ".": "\\."
})


def __make_params_serialized(json_data: dict):

    def encode(data):
        if isinstance(data, dict):
            return encode_dict(data)
        elif isinstance(data, list):
            return encode_list(data)
        else:
            return escape(data)

    def encode_dict(data: dict):
        result = ".".join(_encode_dict(data))
        return "{" + result + "}"

    def _encode_dict(data: dict):
        for key in sorted(data.keys()):
            yield key
            yield encode(data[key])

    def encode_list(data: list):
        result = ".".join(_encode_list(data))
        return f"[" + result + "]"

    def _encode_list(data: list):
        for item in data:
            yield encode(item)

    def escape(data):
        if data is None:
            return "\\0"

        data = str(data)
        return data.translate(translator)

    return ".".join(_encode_dict(json_data))


def serialize(params: dict) -> bytes:
    """
    Serialized params of an original JSON request starting with `icx_sendTransaction`
    to generate a message hash for a signature.

    :param params: params in a original JSON request for transaction.
    :return: serialized params.
    For example, data like `icx_sendTransaction.<key1>.<value1>.<key2>.<value2>` is converted to bytes.
    """
    copy_tx = deepcopy(params)
    key_name_for_tx_hash = __get_key_name_for_tx_hash(params)

    if key_name_for_tx_hash in copy_tx:
        del copy_tx[key_name_for_tx_hash]

    if 'signature' in copy_tx:
        del copy_tx['signature']

    partial_serialized_params = __make_params_serialized(copy_tx)
    return f"icx_sendTransaction.{partial_serialized_params}".encode()


def generate_message(params: dict):
    """
    Generates transaction's message hash from params in request for transaction.

    :param params: params in request for transaction.
    :return: the 256 bit hash digest of a message. Hexadecimal encoded.
    """
    bytes_message_hash = serialize(params)
    return sha3_256(bytes_message_hash).hexdigest()


def __get_key_name_for_tx_hash(params):
    if __get_tx_version(params) == hex(2):
        return "tx_hash"
    else:
        return None


def __get_tx_version(params):
    if 'version' not in params:
        return hex(2)
    else:
        return params['version']



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from iconsdk.utils.convert_type import convert_hex_str_to_int, convert_hex_str_to_bytes
from iconsdk.utils.hexadecimal import add_0x_prefix, is_0x_prefixed, remove_0x_prefix
from iconsdk.utils.type import is_integer


def convert_transaction(transaction: dict):
    """
    Convert transaction data into the right format.
    It supports data about a transaction made not only from JSON RPC V3 but also from V2.

    [what to do on the method 'convert_common_data_on_transaction']
    1. Fields as value, fee, nid, stepLimit, timestamp, and nonce have to be converted to an integer.
    2. Fields as timestamp and nonce have a different type of the value by the version as V2 or V3.
        - If the version V3, the data type is hex string prefixed with '0x'.
        - If the version V2, the data type is a string but the meaning is an integer.
    3. The field 'method' has to be removed.
    4. The field 'version' has to be added number 2 if the version is 2 or if 3, has to be converted to an integer.
    5. If the field 'dataType' is 'deploy', the field 'content' has to be converted to bytes.
       Or if 'message', the field 'data' has to be converted to an integer.

    [Added]
    6. Fields as 'txIndex' and 'blockHeight' have to be converted to an integer.

    :param transaction: data about the single transaction
    """
    convert_common_data_on_transaction(transaction)

    # List of Fields which have to be converted to int
    int_fields = ["txIndex", "blockHeight"]

    for int_field in int_fields:
        if int_field in transaction:
            transaction[int_field] = convert_hex_str_to_int(transaction[int_field])


def convert_block(data: dict):
    """
    Convert block data into the right format.
    It supports data about a block made not only from JSON RPC V3 but also from V2.

    1. If the genesis block, don't convert data because we can not know what data format it is.
    2. Transaction list on a block has to be converted by the method 'convert_common_data_on_transaction' for each transaction.
       You can check what to do on the method in the docstring of the function 'convert_transaction'.
    3. The field name 'tx_hash' on the transaction made from JSON RPC V2 has to be converted to 'txHash'.
       Furthermore, the value of the field has to be prefixed with '0x'.

    :param data: data about the block
    """
    # Genesis block
    if data["height"] == 0:
        return data

    for transaction in data["confirmed_transaction_list"]:
        convert_common_data_on_transaction(transaction)

        if "tx_hash" in transaction:
            transaction["txHash"] = add_0x_prefix(transaction["tx_hash"])
            del transaction["tx_hash"]


def convert_common_data_on_transaction(transaction: dict):
    """
    Convert common fields in the transaction such as value, fee, nid, stepLimit, timestamp, nonce, method, version, data.
    Used in validating a transaction list in a block or validating a single transaction.

    1. Fields such as value, fee, nid, stepLimit, timestamp, and nonce have to be converted to an integer.
    2. Fields such as timestamp and nonce have a different type of the value by the version as V2 or V3.
        - If the version V3, the data type is hex string prefixed with '0x'.
        - If the version V2, the data type is a string but the meaning is an integer.
    3. The field 'method' has to be removed.
    4. The field 'version' has to be added number 2 if the version is 2 or if 3, has to be converted to an integer.
    5. If the field 'dataType' is 'deploy', the field 'content' has to be converted to bytes.
       Or if 'message', the field 'data' has to be converted to an integer.

    :param transaction: data about the single transaction
    """

    # List of Fields which have to be converted to int
    int_fields = ["value", "fee", "nid", "stepLimit"]

    for int_field in int_fields:
        if int_field in transaction:
            transaction[int_field] = convert_hex_str_to_int(transaction[int_field])

    odd_fields = ["timestamp", "nonce"]

    for odd_field in odd_fields:
        if odd_field in transaction:
            if is_integer(transaction[odd_field]):
                pass
            elif is_0x_prefixed(transaction[odd_field]):
                transaction[odd_field] = convert_hex_str_to_int(transaction[odd_field])
            else:
                transaction[odd_field] = int(transaction[odd_field])

    if "method" in transaction:
        del transaction["method"]

    if "version" in transaction and convert_hex_str_to_int(transaction["version"]) >= 3:
        transaction["version"] = convert_hex_str_to_int(transaction["version"])
    else:
        transaction["version"] = 2

    if "dataType" in transaction:
        if transaction["dataType"] == "deploy":
            transaction["data"]["content"] = convert_hex_str_to_bytes(transaction["data"]["content"])
        elif transaction["dataType"] == "message":
            transaction["data"] = bytearray.fromhex(remove_0x_prefix(transaction["data"])).decode()


def convert_transaction_result(data: dict):
    """
    Convert transaction result data into the right format.
    It supports data about a transaction made not only from JSON RPC V3 but also from V2.

    1. Fields such as status, blockHeight, txIndex, stepUsed, stepPrice, cumulativeStepUsed have to be converted to an integer.
    2. The field 'logsBloom' has to be converted to bytes.

    :param data: data about the transaction result
    """
    # Only for the transaction made with JSON RPC V2 successfully does not have the property 'status'
    if "status" not in data and "code" in data and data["code"] == 0:
        data["status"] = 1
        del data["code"]
        return

    # List of Fields which have to be converted to int
    int_fields = ["status", "blockHeight", "txIndex", "stepUsed", "stepPrice", "cumulativeStepUsed"]

    for int_field in int_fields:
        if int_field in data:
            data[int_field] = convert_hex_str_to_int(data[int_field])

    if "logsBloom" in data:
        data["logsBloom"] = convert_hex_str_to_bytes(data["logsBloom"])






# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from hashlib import sha3_256
from abc import ABCMeta, abstractmethod
from secp256k1 import PrivateKey
from iconsdk.utils.validation import is_keystore_file
from iconsdk.exception import KeyStoreException, DataTypeException
from eth_keyfile import create_keyfile_json, extract_key_from_keyfile
from multipledispatch import dispatch
from iconsdk.utils import store_keystore_file_on_the_path
from iconsdk.libs.signer import sign


class Wallet(metaclass=ABCMeta):
    """An interface `Wallet` has 2 abstract methods, `get_address()` and `sign(hash: str)`."""

    @abstractmethod
    def get_address(self) -> str:
        """Gets a wallet address of wallet which starts with 'hx'.

        :return address: A wallet address
        """
        raise NotImplementedError("Wallet must implement this method")

    @abstractmethod
    def sign(self, data: bytes) -> str:
        """Generates signature from input data which is transaction data

        :param data: data to be signed
        :return signature: signature made from input
        """
        raise NotImplementedError("Wallet implement this method")


class KeyWallet(Wallet):
    """KeyWallet class implements Wallet."""

    def __init__(self, private_key_object):
        self.__bytes_private_key = private_key_object.private_key
        self.bytes_public_key = get_public_key(private_key_object)
        self.address = get_address(self.bytes_public_key)

    @staticmethod
    def create():
        """Generates an instance of Wallet without a specific private key.

        :return: An instance of Wallet class.
        """
        private_key_object = PrivateKey()
        wallet = KeyWallet(private_key_object)
        return wallet

    @staticmethod
    @dispatch(bytes)
    def load(private_key: bytes):
        """Loads a wallet from a private key and generates an instance of Wallet.

        :param private_key: private key in bytes
        :return: An instance of Wallet class.
        """
        try:
            private_key_object = PrivateKey(private_key)
            wallet = KeyWallet(private_key_object)
            return wallet
        except TypeError:
            raise DataTypeException("Private key is invalid.")

    @staticmethod
    @dispatch(str, str)
    def load(file_path, password):
        """Loads a wallet from a keystore file with your password and generates an instance of Wallet.

        :param file_path: File path of the keystore file. type(str)
        :param password:
            Password for the keystore file.
            It must include alphabet character, number, and special character.
        :return: An instance of Wallet class.
        """
        try:
            with open(file_path, 'rb') as file:
                bytes_private_key = extract_key_from_keyfile(file, bytes(password, 'utf-8'))
                private_key_object = PrivateKey(bytes_private_key)
                wallet = KeyWallet(private_key_object)
                return wallet
        except FileNotFoundError:
            raise KeyStoreException("File is not found.")
        except ValueError:
            raise KeyStoreException("Password is wrong.")
        except Exception as e:
            raise KeyStoreException(f'keystore file error.{e}')

    def store(self, file_path, password):
        """Stores data of an instance of a derived wallet class on the file path with your password.

        :param file_path: File path of the keystore file. type(str)
        :param password:
            Password for the keystore file. Password must include alphabet character, number, and special character.
            type(str)
        """
        try:
            key_store_contents = create_keyfile_json(
                self.__bytes_private_key,
                bytes(password, 'utf-8'),
                iterations=16384,
                kdf="scrypt"
            )
            key_store_contents['address'] = self.get_address()
            key_store_contents['coinType'] = 'icx'

            # validate the  contents of a keystore file.
            if is_keystore_file(key_store_contents):
                json_string_keystore_data = json.dumps(key_store_contents)
                store_keystore_file_on_the_path(file_path, json_string_keystore_data)
        except FileExistsError:
            raise KeyStoreException("File already exists.")
        except PermissionError:
            raise KeyStoreException("Not enough permission.")
        except FileNotFoundError:
            raise KeyStoreException("File not found.")
        except IsADirectoryError:
            raise KeyStoreException("Directory is invalid.")

    def get_private_key(self) -> str:
        """Returns the private key of the wallet.

        :return a private_key in hexadecimal.
        """
        return self.__bytes_private_key.hex()

    def get_address(self) -> str:
        """Returns an EOA address.

        :return address: An EOA address
        """
        return self.address

    def sign(self, data: bytes) -> bytes:
        """Generates signature from input data which is transaction data

        :param data: data to be signed
        :return signature: signature made from input
        """
        return sign(data, self.__bytes_private_key)


def get_public_key(private_key_object):
    return private_key_object.pubkey.serialize(compressed=False)


def get_address(bytes_public_key):
    return f'hx{sha3_256(bytes_public_key[1:]).digest()[-20:].hex()}'



# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import path
from functools import wraps
from logging import StreamHandler, Formatter
from time import time
from iconsdk.utils.hexadecimal import add_0x_prefix


def store_keystore_file_on_the_path(file_path, json_string):
    """Stores a created keystore string data which is JSON format on the file path.

    :param file_path: The path where the file will be saved. type(str)
    :param json_string: Contents of the keystore.
    """
    if path.isfile(file_path):
        raise FileExistsError

    with open(file_path, 'wt') as f:
        f.write(json_string)


def apply_to_return_value(callback):
    def outer(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            return callback(fn(*args, **kwargs))

        return inner
    return outer


to_dict = apply_to_return_value(dict)


def set_logger(logger, level):
    handler = StreamHandler()
    formatter = Formatter(
        '%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(level)


def get_timestamp():
    """Get epoch time"""
    return add_0x_prefix(hex(int(time() * 10 ** 6)))





# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import requests
from logging import getLogger
from itertools import count
from iconsdk.utils import to_dict
from iconsdk.providers.provider import Provider
from iconsdk.utils import set_logger
from iconsdk.exception import JSONRPCException


class HTTPProvider(Provider):
    """
    The HTTPProvider takes the full URI where the server can be found.
    For local development this would be something like http://localhost:9000.
    """

    endpoint_uri = None
    _request_kwargs = None

    logger = getLogger("HTTPProvider")

    # No need to use logging, remove the line.
    # set_logger(logger, 'DEBUG')

    def __init__(self, endpoint_uri=None, request_kwargs=None):
        self.logger.debug("Init HTTP Provider")
        self.request_counter = count()
        if endpoint_uri is None:
            raise JSONRPCException("Not found endpoint uri.")
        else:
            self.endpoint_uri = endpoint_uri
        self._request_kwargs = request_kwargs or {}

    def __str__(self):
        return "RPC connection {0}".format(self.endpoint_uri)

    @to_dict
    def get_request_kwargs(self):
        if 'headers' not in self._request_kwargs:
            yield 'headers', {'Content-Type': 'application/json'}
        for key, value in self._request_kwargs.items():
            yield key, value

    @staticmethod
    def make_post_request(endpoint_uri, data, **kwargs):
        kwargs.setdefault('timeout', 10)
        with requests.Session() as session:
            response = session.post(url=endpoint_uri, data=json.dumps(data), **kwargs)
        return json.loads(response.content)

    def make_request(self, method, params=None):
        rpc_dict = {
            'jsonrpc': '2.0',
            'method': method,
            'id': 1234
        }

        if params:
            rpc_dict['params'] = params

        self.logger.debug("request HTTP\nURI: %s\nMethod: %s\nData: %s",
                              self.endpoint_uri, method, rpc_dict)

        response = self.make_post_request(self.endpoint_uri, rpc_dict, **self.get_request_kwargs())
        self.logger.debug("response HTTP\nResponse:%s", response)
        return self.return_customed_response(response)

    @staticmethod
    def return_customed_response(response):
        try:
            return response["result"]
        except KeyError:
            raise JSONRPCException(response["error"])

    def is_connected(self):
        try:
            self.logger.debug("Connected")
            response = self.make_request('icx_getLastBlock', [])
        except IOError:
            return False
        else:
            return True




# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from iconsdk.utils.validation import (
    is_block_height,
    is_hex_block_hash,
    is_predefined_block_value,
    is_score_address,
    is_wallet_address,
    is_T_HASH
)
from iconsdk.exception import AddressException, DataTypeException
from iconsdk.providers.provider import Provider
from iconsdk.utils.hexadecimal import add_0x_prefix, remove_0x_prefix
from iconsdk.builder.call_builder import Call
from iconsdk.signed_transaction import SignedTransaction
from iconsdk.converter import convert_block, convert_transaction, convert_transaction_result


class IconService:
    """
    The IconService class contains a set of API methods.
    It accepts a HTTPProvider which serves the purpose of 
    connecting to HTTP and HTTPS based JSON-RPC servers.
    """

    def __init__(self, provider: Provider):
        self.__provider = provider

    def get_block(self, value):
        """
        If param is height,
            1. Returns block information by block height
            2. Delegates to icx_getBlockByHeight RPC method

        Or block hash,
            1. Returns block information by block hash
            2. Delegates to icx_getBlockByHash RPC method

        Or string value same as `latest`,
            1. Returns the last block information
            2. Delegates to icx_getLastBlock RPC method

        :param value:
            Integer of a block height
            or hash of a block prefixed with '0x'
            or `latest`
        :return result: Block data
        """
        # by height
        if is_block_height(value):
            params = {'height': add_0x_prefix(hex(value))}
            result = self.__provider.make_request('icx_getBlockByHeight', params)
        # by hash
        elif is_hex_block_hash(value):
            params = {'hash': value}
            result = self.__provider.make_request('icx_getBlockByHash', params)
        # by value
        elif is_predefined_block_value(value):
            result = self.__provider.make_request('icx_getLastBlock')
        else:
            raise DataTypeException("It's unrecognized block reference:{0!r}.".format(value))

        convert_block(result)
        return result

    def get_total_supply(self):
        """
        Returns total ICX coin supply that has been issued
        Delegates to icx_getTotalSupply RPC method

        :return: Total number of ICX coins issued
        """
        result = self.__provider.make_request('icx_getTotalSupply')
        return int(remove_0x_prefix(result), 16)

    def get_balance(self, address: str):
        """
        Returns the ICX balance of the given EOA or SCORE.
        Delegates to icx_getBalance RPC method.

        :param address: An address of EOA or SCORE. type(str)
        :return: Number of ICX coins
        """

        if is_score_address(address) or is_wallet_address(address):
            params = {'address': address}
            result = self.__provider.make_request('icx_getBalance', params)
            return int(remove_0x_prefix(result), 16)
        else:
            raise AddressException("Address is wrong.")

    def get_score_api(self, address: str):
        """
        Returns SCORE's external API list.
        Delegates to icx_getScoreApi RPC method.

        :param address: A SCORE address to be examined
        :return: A list of API methods of the SCORE and its information
        """
        if is_score_address(address):
            params = {'address': address}
            return self.__provider.make_request('icx_getScoreApi', params)
        else:
            raise AddressException("SCORE Address is wrong.")

    def get_transaction_result(self, tx_hash: str):
        """
        Returns the transaction result requested by transaction hash.
        Delegates to icx_getTransactionResult RPC method.

        :param tx_hash: Hash of a transaction prefixed with '0x'
        :return A transaction result object
        """
        if is_T_HASH(tx_hash):
            params = {'txHash': tx_hash}
            result = self.__provider.make_request('icx_getTransactionResult', params)
            convert_transaction_result(result)
            return result
        else:
            raise DataTypeException("This hash value is unrecognized.")

    def get_transaction(self, tx_hash: str):
        """
        Returns the transaction information requested by transaction hash.
        Delegates to icx_getTransactionByHash RPC method.

        :param tx_hash: Transaction hash prefixed with '0x'
        :return: Information about a transaction
        """
        if is_T_HASH(tx_hash):
            params = {'txHash': tx_hash}
            result = self.__provider.make_request('icx_getTransactionByHash', params)
            convert_transaction(result)
            return result
        else:
            raise DataTypeException("This hash value is unrecognized.")

    def call(self, call: object):
        """
        Calls SCORE's external function which is read-only without creating a transaction on Loopchain.
        Delegates to icx_call RPC method.

        :param call: Call object made by CallBuilder
        :return: Values returned by the executed SCORE function
        """
        if isinstance(call, Call):
            params = {
                "from": call.from_,
                "to": call.to,
                "dataType": "call",
                "data": {
                    "method": call.method
                }
            }

            if isinstance(call.params, dict):
                params["data"]["params"] = call.params

            return self.__provider.make_request('icx_call', params)
        else:
            raise DataTypeException("Call object is unrecognized.")

    def send_transaction(self, signed_transaction: SignedTransaction):
        """
        Sends the transaction.
        Delegates to icx_sendTransaction RPC method.

        :param signed_transaction: A signed transaction object
        :return: Transaction hash prefixed with '0x'
        """
        params = signed_transaction.signed_transaction_dict
        return self.__provider.make_request('icx_sendTransaction', params)






# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from enum import IntEnum, unique


@unique
class IconServiceExceptionCode(IntEnum):
    """Result code enumeration"""
    OK = 0
    KEY_STORE_ERROR = 1
    ADDRESS_ERROR = 2
    BALANCE_ERROR = 3
    DATA_TYPE_ERROR = 4
    JSON_RPC_ERROR = 5
    ZIP_MEMORY_ERROR = 6

    def __str__(self) -> str:
        return str(self.name).capitalize().replace('_', ' ')


class IconServiceBaseException(BaseException):

    def __init__(self, message: Optional[str], code: IconServiceExceptionCode = IconServiceExceptionCode.OK):
        if message is None:
            message = str(code)
        self.__message = message
        self.__code = code

    @property
    def message(self):
        return self.__message

    @property
    def code(self):
        return self.__code

    def __str__(self):
        return f'{self.message} ({str(self.code)})'


class KeyStoreException(IconServiceBaseException):
    """"Error when making or loading a keystore file."""
    def __init__(self, message: Optional[str]):
        super().__init__(message, IconServiceExceptionCode.KEY_STORE_ERROR)


class AddressException(IconServiceBaseException):
    """Error when having an invalid address."""
    def __init__(self, message: Optional[str]):
        super().__init__(message, IconServiceExceptionCode.ADDRESS_ERROR)


class BalanceException(IconServiceBaseException):
    """Error when having an invalid balance."""
    def __init__(self, message: Optional[str]):
        super().__init__(message, IconServiceExceptionCode.BALANCE_ERROR)


class DataTypeException(IconServiceBaseException):
    """Error when data type is invalid."""
    def __init__(self, message: Optional[str]):
        super().__init__(message, IconServiceExceptionCode.DATA_TYPE_ERROR)


class JSONRPCException(IconServiceBaseException):
    """Error when get JSON-RPC Error Response."""
    def __init__(self, message: Optional[str]):
        super().__init__(message, IconServiceExceptionCode.JSON_RPC_ERROR)


class ZipException(IconServiceBaseException):
    """"Error while write zip in memory"""
    def __init__(self, message: Optional[str]):
        super().__init__(message, IconServiceExceptionCode.ZIP_MEMORY_ERROR)




# -*- coding: utf-8 -*-
# Copyright 2018 ICON Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from os import environ, path

version = environ.get('VERSION')

if version is None:
    with open(path.join('.', 'VERSION')) as version_file:
        version = version_file.read().strip()

with open("README.md", 'r') as f:
    long_description = f.read()

setup(
   name='iconsdk',
   version=version,
   description='ICON SDK for Python is a collection of libraries which allow you to interact '
               'with a local or remote Loopchain node, using an HTTP connection. ',
   long_description=long_description,
   long_description_content_type='text/markdown',
   author='ICON foundation',
   author_email='foo@icon.foundation',
   url='https://github.com/icon-project/icon-sdk-python',
   packages=find_packages(exclude=['tests*']),
   test_suite='tests',
   install_requires=[
        'eth-keyfile==0.5.1',
        'secp256k1==0.13.2',
        'multipledispatch==0.5.0',
        'requests>=2.20.0'
   ],
   license='Apache License 2.0',
   classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6'
    ]
)
